from .base import *

'''
Construtor del hamiltoniano de fermi-hubbard
'''
class vqe_fermihubbard(vqe_base):
    def __init__(self, params, lat):
        self.qubits = params["sites"]*2
        fermi_sentence = 0.0

        if lat['lattice'] == 'custom':
            pass
        else:
            x,y = lat['size']
            lattice_edge, lattice_node = lattice(lat)
            #-t term
            hopping = -params["hopping"]
            fermi_hopping = 0.0
            for pair in lattice_edge:
                p1, p2 = pair
                fermi_hopping +=  FermiC(2*(y*p1[0] + p1[1]))*FermiA(2*(y*p2[0] + p2[1])) + FermiC(2*(y*p2[0] + p2[1]))*FermiA(2*(y*p1[0] + p1[1]))
                fermi_hopping +=  FermiC(2*(y*p1[0] + p1[1])+1)*FermiA(2*(y*p2[0] + p2[1])+1) + FermiC(2*(y*p2[0] + p2[1])+1)*FermiA(2*(y*p1[0] + p1[1])+1)

            fermi_sentence = hopping*fermi_hopping

            #U term
            if 'U' in params:
                Upotential = params["U"]
                fermi_U = 0.0
                for node in lattice_node:
                    p1, p2 = node
                    fermi_U += FermiC(y*p1 + 2*p2)*FermiA(y*p1 + 2*p2)*FermiC(y*p1 + 2*p2+1)*FermiA(y*p1 + 2*p2+1)
                fermi_sentence += Upotential*fermi_U

            #E term
            if 'E' in params:
                Efield = params["E"]
                fermi_E = 0.0
                for node in lattice_node:
                    p1, p2 = node
                    fermi_E += FermiC(y*p1 + 2*p2)*FermiA(y*p1 + 2*p2)
                    fermi_E += FermiC(y*p1 + 2*p2+1)*FermiA(y*p1 + 2*p2+1)
                fermi_sentence += Efield*fermi_E

            #V term
            if 'V' in params:
                Vpotencial = params["V"]
                fermi_V = 0.0
                for pair in lattice_edge:
                    p1, p2 = pair
                    n_i = FermiC(2*(y*p1[0] + p1[1]))*FermiA(2*(y*p1[0] + p1[1])) + FermiC(2*(y*p1[0] + p1[1]) +1)*FermiA(2*(y*p1[0] + p1[1]) +1)
                    n_j = FermiC(2*(y*p2[0] + p2[1]))*FermiA(2*(y*p2[0] + p2[1])) + FermiC(2*(y*p2[0] + p2[1]) +1)*FermiA(2*(y*p2[0] + p2[1]) +1)
                    fermi_V += n_i*n_j
                fermi_sentence += Vpotencial*fermi_V


        coeff, terms = qml.jordan_wigner( fermi_sentence, ps=True ).hamiltonian().terms()

        to_delete = []
        for i,c in enumerate(coeff):
            if c==0.0:
                to_delete.append(i)
            if isinstance(terms[i], qml.Identity):
                terms[i] = qml.Identity(wires=[0])
        
        to_delete = -np.sort(-np.array(to_delete))
        for index in to_delete:
            coeff.pop(index)
            terms.pop(index)

        self.hamiltonian = qml.Hamiltonian(np.real(np.array(coeff)), terms)
        return


import pyscf
from pyscf import gto, scf, ci, cc
import numpy as np
import pennylane as qml

AngtoAU = 1.8897259886 #a.u

def remove_special_characters(input_string):
    result = ''.join(char for char in input_string if char.isalnum() or char.isspace())
    return result

def CISD_state(elements, coordinates, params):
    atom = []
    for i in range( len(elements) ):
        atom.append( [elements[i], (( coordinates[3*i], coordinates[3*i+1], coordinates[3*i+2] )) ] )
   
    basis = remove_special_characters( params["basis"] )
    mol = gto.M(atom=atom, charge=params["charge"], basis=basis, unit="B")
    myhf = scf.RHF(mol).run()
    myci = ci.CISD(myhf).run()
    wf_cisd = qml.qchem.import_state(myci, tol=1e-6)
    return wf_cisd.real



def CCSD_state(elements, coordinates, params):
    atom = []
    for i in range( len(elements) ):
        atom.append( [elements[i], (( coordinates[3*i], coordinates[3*i+1], coordinates[3*i+2] )) ] )
   
    mol = gto.M(atom=atom, charge=params["charge"], unit="B")
    myhf = scf.RHF(mol).run()
    mycc = cc.CCSD(myhf).run()

    wf_cisd = qml.qchem.import_state(mycc, tol=1e-6)
    return wf_cisd.real


def HF_state():


    return



from .base import *

class adap_molecular(adap_base):
    mapping= 'jordan_wigner'
    charge= 0
    mult= 1
    basis='sto-3g'
    method='dhf'
    active_electrons = None
    active_orbitals = None

    def __init__(self, symbols, coordinates, params= None):
        if 'mapping' in params:
            self.mapping = params['mapping']
            
        if 'charge' in params:
            self.charge = params['charge']

        if 'mult' in params:
            self.mult = params['mult']

        if 'basis' in params:
            self.basis = params['basis']
        
        if 'method' in params:
            self.method = params['method']
            
        if 'active_electrons' in params:
            self.active_electrons = params['active_electrons']

        if 'active_orbitals' in params:
            self.active_orbitals = params['active_orbitals']

        self.hamiltonian, self.qubits = qchem.molecular_hamiltonian(
            symbols= symbols,
            coordinates= coordinates,
            mapping= self.mapping,
            charge= self.charge,
            mult= self.mult,
            basis= self.basis,
            method= self.method,
            active_electrons=self.active_electrons, 
            active_orbitals=self.active_orbitals,
            load_data=True)
        return



import pennylane as qml
import jax
import jax.numpy as jnp
import optax
jax.config.update("jax_enable_x64", True)

from pennylane import numpy as np
from quantumsim.optimizers import *

class jax_optimizer():
    maxiter = 100
    theta_optimizer = None
    x_optimizer = None
    tol = 1e-6
    number = 0
    begin_state= None

    get_energy = False
    get_params = False

    def __init__(self, params):
        self.number = params["number"]

        if 'tol' in params:
            self.tol = params["tol"]
        
        if 'maxiter' in params:
            self.maxiter = params["maxiter"]
        
        if 'theta' in params:
            if params['theta'][0] == "adam":
                self.theta_optimizer = optax.adam(learning_rate=params['theta'][1])
            elif params['theta'][0] == "adagrad":
                self.theta_optimizer = optax.adagrad(learning_rate=params['theta'][1])

        if 'x' in params:
            if params['theta'][0] == "adam":
                self.x_optimizer = optax.adam(learning_rate=params['theta'][1])
            elif params['theta'][0] == "adagrad":
                self.x_optimizer = optax.adagrad(learning_rate=params['theta'][1])

        self.begin_state = np.random.random( size=self.number )*(np.pi/180.0)


    def VQE(self, cost_fn):
        theta = jnp.array( self.begin_state )
        opt_state = self.theta_optimizer.init(theta)
        
        energy_evol = [cost_fn(theta)]
        theta_evol = [theta]

        for n in range(self.maxiter):

            grads = jax.grad(cost_fn)(theta)
            updates, opt_state = self.theta_optimizer.update(grads, opt_state)
            theta = optax.apply_updates(theta, updates)

            energy_evol.append(cost_fn(theta))
            theta_evol.append(theta)

            conv = jnp.abs(energy_evol[-1] - energy_evol[-2])
            if conv <= self.tol:
                break
        
        if self.get_energy == True and self.get_params == True:
            return energy_evol, theta_evol
        else:
            if self.get_energy == True:
                return energy_evol, theta_evol[-1]
            elif self.get_params== True:
                return energy_evol[-1], theta_evol
            else:
                return energy_evol[-1], theta_evol[-1]
    

    def OS(self, cost_fn, x, grad):
        theta = self.begin_state
        opt_state_theta = self.theta_optimizer.init(theta)

        opt_state_x = self.x_optimizer.init(x)
        theta_evol = [ jnp.concatenate( (x, theta) ) ]
        energy_evol = [cost_fn(theta, x)]

        for n in range(self.maxiter):
            grads = jax.grad(cost_fn)(theta, x)
            updates_theta, opt_state_theta = self.theta_optimizer.update(grads, opt_state_theta)
            theta = optax.apply_updates(theta, updates_theta)


            updates_x, opt_state_x = self.x_optimizer.update(grads, opt_state_x)
            x = optax.apply_updates( x, updates_x )

            theta_evol.append( jnp.concatenate( (x, theta) ) )
            energy_evol.append( cost_fn( theta, x ) )
            if jnp.max( grad(theta, x) ) <= self.tol:
                break
        
        if self.get_energy == True and self.get_params == True:
            return energy_evol, theta_evol
        else:
            if self.get_energy == True:
                return energy_evol, theta_evol[-1]
            elif self.get_params== True:
                return energy_evol[-1], theta_evol
            else:
                return energy_evol[-1], theta_evol[-1]






import pennylane as qml
from pennylane import numpy as np
from .base import *

class uccds_ansatz(base_ansatz):
    begin_state = None
    singles = None
    doubles = None

    def set_exitations(self, electrons, sz):
        singles, doubles = qml.qchem.excitations(electrons, self.qubits, delta_sz=sz)
        self.singles, self.doubles = qml.qchem.excitations_to_wires(singles, doubles)
    

    def set_state(self, electrons, sz):
        self.begin_state = qml.qchem.hf_state(electrons=electrons, orbitals=self.qubits)


    def circuit(self, theta, obs):
        qml.UCCSD(theta, range(self.qubits), self.singles, self.doubles, self.begin_state)
        return qml.expval(obs)
    

    def circuit_state(self, theta):
        qml.UCCSD(theta, range(self.qubits), self.singles, self.doubles, self.begin_state)
        return qml.state()



from quantumsim.ansatz import *
from quantumsim.optimizers import *
from pennylane import qchem
from pennylane import numpy as np

class structure_molecular():
    symbols = None
    coordinates = None
    begin_state = None

    mapping= 'jordan_wigner'
    charge= 0
    mult= 1
    basis='sto-3g'
    method='dhf'
    active_electrons = None
    active_orbitals = None

    node = None
    interface = None

    def __init__(self, symbols, coordinates, params= None):
        self.symbols = symbols
        self.coordinates = coordinates
        
        if 'mapping' in params:
            self.mapping = params['mapping']
            
        if 'charge' in params:
            self.charge = params['charge']

        if 'mult' in params:
            self.mult = params['mult']

        if 'basis' in params:
            self.basis = params['basis']

        if 'method' in params:
            self.method = params['method']
            
        if 'active_electrons' in params:
            self.active_electrons = params['active_electrons']

        if 'active_orbitals' in params:
            self.active_orbitals = params['active_orbitals']

        _, self.qubits = qchem.molecular_hamiltonian(
            symbols= symbols,
            coordinates= coordinates,
            mapping= self.mapping,
            charge= self.charge,
            mult= self.mult,
            basis= self.basis,
            method= self.method,
            active_electrons=self.active_electrons,
            active_orbitals=self.active_orbitals,
            load_data=True)
        
        self.begin_state = qml.qchem.hf_state(self.active_electrons, self.qubits)
        return
    
    def set_node(self, node, interface) -> None:
        self.node = node
        self.interface = interface
        return
    
    
    def grad_x(self, theta, x):
        delta = 0.01
        n = len(x)
        shift = np.eye( n ) * 0.5 * delta
        grad = [ self.node( theta=theta, obs=((self.H(x + shift[i]) - self.H(x - shift[i])) / delta) ) for i in range(n)]
        return np.array(grad)
    

    def H(self, x):
        H, q = qchem.molecular_hamiltonian(symbols= self.symbols,
            coordinates= x,
            mapping= self.mapping,
            charge= self.charge,
            mult= self.mult,
            basis= self.basis,
            method= self.method,
            active_electrons=self.active_electrons,
            active_orbitals=self.active_orbitals,
            load_data=True)
        return H    

    def cost_function(self, theta, x):
        hamiltonian = self.H(x)
        result = self.node( theta=theta, obs=hamiltonian )
        return result


import pennylane as qml
from pennylane import numpy as np
from quantumsim.optimizers import *

class TookTooManyIters(Warning):
    pass


'''
The adaptative optimizer use a pool of uccsd operators (single and double operators)
This optimizer returns the minimum circuit's depth with the minimun number
of parameters.

'''
class adap_optimizer():
    maxiter = 20
    optimizer = None
    operator_pool = None
    tol = 1e-6

    def __init__(self, params):
        
        if 'maxiter' in params:
            self.maxiter = params["maxiter"]

        if 'tol' in params:
            self.tol = params["tol"]

        if 'theta' in params:
            self.optimizer = qml.GradientDescentOptimizer(stepsize=params['theta'][1])
        
        if 'electrons' in params and 'qubits' in params and 'sz' in params:
            singles, doubles = qml.qchem.excitations(params['electrons'], params['qubits'], delta_sz= params['sz'])
            #print(len(singles)+len(doubles))
            self.operator_pool = [ doubles, singles ]



    def MinimumCircuit(self, cost_fn):
        params_doubles = [0.0]*len(self.operator_pool[0])
        circuit_gradient = qml.grad(cost_fn, argnum=0)
        grad_doubles = circuit_gradient(params_doubles, excitations=self.operator_pool[0])
        doubles_select = [self.operator_pool[0][i] for i in range(len(self.operator_pool[0])) if abs(grad_doubles[i]) > self.tol]
        params_doubles = np.zeros(len(doubles_select), requires_grad=True)
        if len(params_doubles) != 0:
            for n in range(self.maxiter):
                params_doubles = self.optimizer.step(cost_fn, params_doubles, excitations=doubles_select)


        circuit_gradient = qml.grad(cost_fn, argnum=0)
        params_single = [0.0] * len(self.operator_pool[1])   
        grad_singles = circuit_gradient(
            params_single,
            excitations=self.operator_pool[1],
            gates_select=doubles_select,
            params_select=params_doubles
        )
        singles_select = [self.operator_pool[1][i] for i in range(len(self.operator_pool[1])) if abs(grad_singles[i]) > self.tol]
        return singles_select, doubles_select
    


from .base import *

'''
Construtor del hamiltoniano de espines
'''
class vqe_spin(vqe_base):
    def __init__(self, params, lat):
        self.qubits = params['sites']
        terms = []
        coeff = []

        if lat['lattice'] == 'custom':
            pass
        else:
            x,y = lat['size']
            lattice_edge, lattice_node = lattice(lat)

            #J term
            for pair in lattice_edge:
                termz = qml.PauliZ(wires=[ x*pair[0][0] + pair[0][1]])@qml.PauliZ(wires=[x*pair[1][0] + pair[1][1]])
                termx = qml.PauliX(wires=[ x*pair[0][0] + pair[0][1]])@qml.PauliX(wires=[x*pair[1][0] + pair[1][1]])
                termy = qml.PauliY(wires=[ x*pair[0][0] + pair[0][1]])@qml.PauliY(wires=[x*pair[1][0] + pair[1][1]])

                terms.extend( [termx, termy, termz] )
                coeff.extend( [-params["Jx"][0]/4.0, -params["Jy"][1]/4.0, -params["Jz"][2]/4.0] )

            #h term
            if 'h' in params:
                for pair in lattice_node:
                    termz = qml.PauliZ(wires=[ x*pair[0] + pair[1]])
                    termz = qml.PauliX(wires=[ x*pair[0] + pair[1]])
                    termz = qml.PauliY(wires=[ x*pair[0] + pair[1]])

                    terms.extend( [termx, termy, termz] )
                    coeff.extend( [params["h"][0]/2.0, params["h"][1]/2.0, params["h"][2]/2.0] )
        
        to_delete = []
        for i,c in enumerate(coeff):
            if c==0.0:
                to_delete.append(i)
        
        to_delete = -np.sort(-np.array(to_delete))
        for index in to_delete:
            coeff.pop(index)
            terms.pop(index)

        self.hamiltonian = qml.Hamiltonian(np.real(np.array(coeff)), terms)
        return
    



import pennylane as qml
from pennylane import numpy as np
from quantumsim.optimizers import *

class gradiend_optimizer():
    maxiter = 100
    theta_optimizer = None
    x_optimizer = None
    tensor_metric = None
    tol = 1e-6
    number = 0
    begin_state= None

    get_energy = False
    get_params = False

    def __init__(self, params):
        self.number = params["number"]

        if 'tol' in params:
            self.tol = params["tol"]
        
        if 'maxiter' in params:
            self.maxiter = params["maxiter"]

        if 'theta' in params:
            if params['theta'][0] == "generic":
                self.theta_optimizer = qml.GradientDescentOptimizer(stepsize=params['theta'][1])
            elif params['theta'][0] == "adam":
                self.theta_optimizer = qml.AdamOptimizer(stepsize=params['theta'][1])
            elif params['theta'][0] == "adagrad":
                self.theta_optimizer = qml.AdagradOptimizer(stepsize=params['theta'][1])

        if 'x' in params:
            if params['x'][0] == "generic":
                self.x_optimizer = qml.GradientDescentOptimizer(stepsize=params['x'][1])
            elif params['x'][0] == "adam":
                self.x_optimizer = qml.AdamOptimizer(stepsize=params['x'][1])
            elif params['x'][0] == "adagrad":
                self.x_optimizer = qml.AdagradOptimizer(stepsize=params['x'][1])
        
        self.begin_state = np.random.random( size=self.number )*(np.pi/180.0)


    def VQE(self, cost_function):
        theta = self.begin_state
        energy_evol = [cost_function(theta)]
        theta_evol = [theta]

        for n in range(self.maxiter):
            theta.requires_grad = True
            theta = self.theta_optimizer.step(cost_function, theta)

            energy_evol.append(cost_function(theta))
            theta_evol.append(theta)

            prev_energy = energy_evol[-2]
            conv = np.abs(energy_evol[-1] - prev_energy)
            if conv <= self.tol:
                break
        
        if self.get_energy == True and self.get_params == True:
            return energy_evol, theta_evol
        else:
            if self.get_energy == True:
                return energy_evol, theta_evol[-1]
            elif self.get_params== True:
                return energy_evol[-1], theta_evol
            else:
                return energy_evol[-1], theta_evol[-1]
           


    def OS(self, cost_function, x, grad):
        theta = self.begin_state
        theta_evol = [ np.concatenate( (x, theta)) ]
        energy_evol = [ cost_function( theta, x ) ]

        for _ in range(self.maxiter):
            theta.requires_grad = True
            x.requires_grad = False
            theta, _ = self.theta_optimizer.step(cost_function, theta, x)

            x.requires_grad = True
            theta.requires_grad = False
            _, x = self.x_optimizer.step(cost_function, theta, x, grad_fn=grad)


            theta_evol.append( np.concatenate( (x, theta) ) )
            energy_evol.append( cost_function( theta,x ) )

            print(energy_evol[-1])
            if np.max( grad(theta, x) ) <= self.tol:
                break
        
        if self.get_energy == True and self.get_params == True:
            return energy_evol, theta_evol
        else:
            if self.get_energy == True:
                return energy_evol, theta_evol[-1]
            elif self.get_params== True:
                return energy_evol[-1], theta_evol
            else:
                return energy_evol[-1], theta_evol[-1]
    
    
    #def VQD(self, cost_function, overlap_cost_function, k):
    #    previous_theta = []
    #    energy_final = [] 

    #    for i in range(k):
    #        def cost_aux(x): 
    #            result = cost_function(x) 
    #            for previous in previous_theta:
    #                result += 10*overlap_cost_function(x, previous)
    #            return result
        
    #        print("state ", i+1)
    #        self.nit = 0

            
    #        theta = self.begin_state
    #        energy = [cost_function(theta)]
    #        theta_evol = [theta]
    #        for _ in range(self.maxiter):
    #            theta.requires_grad = True
    #            theta = self.theta_optimizer.step(cost_aux, theta)
    #            energy.append(cost_function(theta))
    #            theta_evol.append(theta)
    #            prev_energy = energy[len(energy)-2]

    #               conv = np.abs(energy[-1] - prev_energy)
    #            if conv <= self.tol:
    #                break
    #        energy_final.append( energy[-1] )
    #        previous_theta.append( theta )
    #    return energy_final, previous_theta



import pennylane as qml
import scipy as sc
import warnings
from pennylane import numpy as np
from quantumsim.optimizers import *

class TookTooManyIters(Warning):
    pass

class scipy_optimizer():
    maxiter = 100
    type_method = "SLSQP"
    tol = 1e-6
    nit = 0
    number = 0
    begin_state= None

    def __init__(self, params):
        self.number = params["number"]

        if 'tol' in params:
            self.tol = params["tol"]
        
        if 'maxiter' in params:
            self.maxiter = params["maxiter"]
        
        if 'type' in params:
            self.type_method = params["type"]

        self.begin_state = np.random.random( size=self.number )*(np.pi/180.0)


    def callback(self, x):
        self.nit += 1
        if self.nit == self.maxiter:
            print("Maximo numero de iteraciones")
            warnings.warn("Terminating optimization: iteration limit reached", TookTooManyIters)


    
    def VQE(self, cost_function):
        energy = []
        theta_evol = []
        self.nit = 0

        def cost_aux(x): 
            result = cost_function(x)
            energy.append(result)
            theta_evol.append(x)
            return result
        
        ops = {'maxiter': self.maxiter}
        theta = sc.optimize.minimize(cost_aux, self.begin_state, method=self.type_method,
                callback=self.callback, tol=self.tol, options=ops)['x']
        return energy, theta
    

    
    def OS(self, cost_function, x):
        energy = []
        theta_evol = []
        self.nit = 0

        def cost_aux(psi): 
            result = cost_function(psi[len(x):], psi[:len(x)])
            energy.append(result)
            theta_evol.append(psi)
            return result
        
        ops = {'maxiter': self.maxiter}
        theta = sc.optimize.minimize(cost_aux, np.concatenate((x,self.begin_state), axis=0), method=self.type_method, 
                callback=self.callback, tol=self.tol, options=ops)['x']
        return energy, theta
    

    #def VQD(self, cost_function, overlap_cost_function, k):
    #    energy = []
    #    previous_theta = []

    #    for i in range(k):  
    #        print("state ", i+1)

    #        def cost_aux(x): 
    #            result = cost_function(x) 
    #            for j, previous in enumerate(previous_theta):
    #                result += 10*overlap_cost_function(x, previous)
    #            return result
        
    #        self.nit = 0
    #        ops = {'maxiter': self.maxiter}
    #        xs = sc.optimize.minimize(cost_aux, self.begin_state, method=self.type_method,
    #                callback=self.callback, tol=self.tol, options=ops)['x']
    #        energy.append( cost_function(xs) )
    #        previous_theta.append(xs)
    #    return energy, previous_theta





import pennylane as qml
from pennylane import numpy as np
from quantumsim.ansatz import *
from quantumsim.lattice import *
from pennylane import qchem
from pennylane import FermiC, FermiA

class adap_base():
    hamiltonian = None
    qubits = 0

    base = ""
    backend = ""
    token = ""
    device= None
    node = None
    begin_state = None

    def set_device(self, params) -> None:
        self.base = params['base']

        ##Maquinas reales
        if self.base == 'qiskit.ibmq':
            if params['backend']:
                self.backend = params['backend']
            else:
                raise Exception("Backend no encontrado")
            if params['token']:
                self.token = params['token']
                self.device= qml.device(self.base, backend=self.backend, 
                    wires=self.qubits,  ibmqx_token= self.token)
            else:
                raise Exception("Token de acceso no encontrado")
        ##Simuladores de qiskit
        elif self.base == "qiskit.aer":
            if params['backend']:
                self.backend = params['backend']
            else:
                raise Exception("Backend no encontrado")
            
            self.device= qml.device(self.base, backend=self.backend, wires=self.qubits)
        ##Simuladores de pennylane
        else:
            self.device= qml.device(self.base, wires=self.qubits)
        return

    def set_node(self, params) -> None:
        if "pattern" in params:
            self.pattern = params["pattern"]
        
        if params['repetitions']:
            self.repetition = params['repetitions']

            
        if params['interface'] == "jax" or params['interface'] == "jax-jit":
            node = qml.QNode(self.circuit, self.device, interface=params['interface'])
            self.node = jax.jit(node)
        else:
            self.node = qml.QNode(self.circuit, self.device, interface=params['interface'])
        return
    
    def set_state(self, electrons):
        self.begin_state = qml.qchem.hf_state(electrons=electrons, orbitals=self.qubits)

    def circuit(self, params, excitations, params_select=None, gates_select=None):
        qml.BasisState(self.begin_state, wires=range(self.qubits))

        if gates_select != None:
            for i, gate in enumerate(gates_select):
                if len(gate) == 4:
                    qml.DoubleExcitation(params_select[i], wires=gate)
                elif len(gate) == 2:
                    qml.SingleExcitation(params_select[i], wires=gate)

        for i, gate in enumerate(excitations):
            if len(gate) == 4:
                qml.DoubleExcitation(params[i], wires=gate)
            elif len(gate) == 2:
                qml.SingleExcitation(params[i], wires=gate)
        return qml.expval(self.hamiltonian)




import pennylane as qml
from pennylane import numpy as np
from .base import *

class kupccgsd_ansatz(base_ansatz):
    sz = 0
    begin_state = None

    
    def set_state(self, electrons, sz):
        self.begin_state = qml.qchem.hf_state(electrons=electrons, orbitals=self.qubits)
        self.sz = sz


    def circuit(self, theta, obs):
        theta = theta.reshape( qml.kUpCCGSD.shape(k=self.repetition, 
            n_wires=self.qubits, delta_sz=self.sz) )
        
        qml.kUpCCGSD(theta, wires=range(self.qubits),
            k=self.repetition, delta_sz=0, init_state=self.begin_state)
        return qml.expval(obs)
    
    def circuit_state(self, theta):
        theta = theta.reshape( qml.kUpCCGSD.shape(k=self.repetition, 
            n_wires=self.qubits, delta_sz=self.sz) )
        
        qml.kUpCCGSD(theta, wires=range(self.qubits),
            k=self.repetition, delta_sz=0, init_state=self.begin_state)
        return qml.state()


import networkx as nx


'''
input:
    params: diccionario con los parametros de la lattice
        -tipo de lattice
        -tamaño: tupla
        -periodicidad
return:
    list of the graph's edges
'''
def lattice(params):

    periodicity = False
    if isinstance(params['bound'], list):
        periodicity = params['bound']
    elif params['bound'] == True:
        periodicity = params['bound']

    if params['lattice'] == 'chain':
        lattice = nx.grid_2d_graph(params['size'][0], params['size'][1], periodicity)

    elif params['lattice'] == 'triangle':
        lattice = nx.triangular_lattice_graph(params['size'][0], params['size'][1], periodicity)

    elif params['lattice'] == 'square':
        lattice = nx.grid_2d_graph(params['size'][0], params['size'][1], periodicity)
    
    elif params['lattice'] == 'hexagon':
        lattice = nx.hexagonal_lattice_graph(params['size'][0], params['size'][1], periodicity)

    return lattice.edges(), lattice.nodes()


from quantumsim.ansatz import *
from quantumsim.lattice import *
from pennylane import qchem
from pennylane import FermiC, FermiA
import itertools

'''
Base del VQE donde se definen las funciones de coste,
y funciones de calculos auxiliares, como las proyecciones
'''
class vqe_base():
    hamiltonian= None
    coeff = None
    qubits = 0
    node = None
    interface = None

    def set_node(self, node, interface) -> None:
        self.node = node
        self.interface = interface
        return
    
    def cost_function(self, theta) -> float:
        result = self.node( theta=theta, obs=self.hamiltonian )
        return result
    

    #Cantidades utiles de saber
    #Theta corresponde al vector de parametros del circuito
    #que construye el estado obtenido de un proceso
    #de minimizacion
    def get_projections(self, theta):
        combos = itertools.product([0, 1], repeat=self.qubits)
        s = [list(c) for c in combos]
        pro = [ qml.Projector(state, wires=range(self.qubits)) for state in s]
        result = [self.node( theta=theta, obs=state) for state in pro]
        return result, s
    
    #Espin total del estado
    def get_totalspinS(self, theta, electrons):
        s_square = qml.qchem.spin2(electrons, self.qubits)
        result = self.node( theta=theta, obs=s_square )
        return result
    
    #Espin total proyectado en S_z del estado
    def get_totalspinSz(self, theta):
        s_z = qml.qchem.spinz(self.qubits)
        result = self.node( theta=theta, obs=s_z )
        return result
    
    #Numero de particulas del estado
    def get_particlenumber(self, theta):
        n = qml.qchem.particle_number(self.qubits)
        result = self.node( theta=theta, obs=n )
        return result




import pennylane as qml
from pennylane import numpy as np
from .base import *

'''
Hardware Efficient ansatz
'''
class he_ansatz(base_ansatz):
    pattern = "chain"
    begin_state = None

    def set_state(self, state):
        self.begin_state = state

    '''
    Funciones para la construccion del ansatz
    '''
    def single_rotation(self, params):
        for i in range(0, self.qubits):
            qml.RY(params[i], wires=[i])
        return

    def non_local_gates(self, flag=0):
        ## Compuertas en orden normal
        if flag == 0:
            if self.pattern == 'chain':
                for i in range(0, self.qubits-1):
                    qml.CNOT(wires=[i, i+1])

            elif self.pattern == 'ring':
                if self.qubits == 2:
                    qml.CNOT(wires=[0, 1])
                else:
                    for i in range(0, self.qubits-1):
                        qml.CNOT(wires=[i, i+1])

                    qml.CNOT(wires=[self.qubits-1, 0])  

        #Compuertas en orden inverso (la operacion inversa)
        else:
            if self.pattern == 'chain':
                for i in range(self.qubits-1,0,-1):
                    qml.CNOT(wires=[i-1, i])

            elif self.pattern == 'ring':
                if self.qubits == 2:
                    qml.CNOT(wires=[0, 1])
                else:
                    qml.CNOT(wires=[self.qubits-1, 0])  
                        
                    for i in range(self.qubits-1,0,-1):
                        qml.CNOT(wires=[i-1, i])
        return
    

    def circuit(self, theta, obs):
        qml.StatePrep(self.begin_state, wires=range(self.qubits))
        rotation_number = self.qubits
        for k in range(0, self.repetition):
            params = theta[k*rotation_number:(k+1)*rotation_number]
            self.single_rotation(params)
            self.non_local_gates(0)
            
        return qml.expval(obs)


    def circuit_state(self, theta):
        qml.StatePrep(self.begin_state, wires=range(self.qubits))
        rotation_number = self.qubits
        for k in range(0, self.repetition):
            params = theta[k*rotation_number:(k+1)*rotation_number]
            self.single_rotation(params)
            self.non_local_gates(0)
        return qml.state()




import pennylane as qml
from pennylane import numpy as np
from .base import *

class custom_ansatz(base_ansatz):
    begin_state = None
    excitations = []
    
    def set_state(self, state):
        self.begin_state = state

    def set_gates(self, singles=None, doubles=None):
        self.excitations = []
        if doubles != None:
            for term in doubles:
                self.excitations.append(term)

        if singles != None:
            for term in singles:
                self.excitations.append(term)
        return
    

    def circuit(self, theta, obs):
        qml.StatePrep(self.begin_state, wires=range(self.qubits))

        for i, gate in enumerate(self.excitations):
            if len(gate) == 4:
                qml.DoubleExcitation(theta[i], wires=gate)
            elif len(gate) == 2:
                qml.SingleExcitation(theta[i], wires=gate)

        return qml.expval(obs)
    
    def circuit_state(self, theta):
        qml.StatePrep(self.begin_state, wires=range(self.qubits))

        for i, gate in enumerate(self.excitations):
            if len(gate) == 4:
                qml.DoubleExcitation(theta[i], wires=gate)
            elif len(gate) == 2:
                qml.SingleExcitation(theta[i], wires=gate)

        return qml.state()
    


import pennylane as qml
from pennylane import numpy as np
import jax

class base_ansatz():
    base = "default.qubit"
    backend = ""
    token = ""
    interface = "autograd"
    diff_method = "best"

    device= None
    node = None
    qubits = 0
    repetition = 0

    def set_device(self, params) -> None:
        self.base = params['base']
        self.qubits = params["qubits"]

        ##Maquinas reales
        if self.base == 'qiskit.ibmq':
            if params['backend']:
                self.backend = params['backend']
            else:
                raise Exception("Backend no encontrado")
            if params['token']:
                self.token = params['token']
                self.device= qml.device(self.base, backend=self.backend, 
                    wires=self.qubits,  ibmqx_token= self.token)
            else:
                raise Exception("Token de acceso no encontrado")
        ##Simuladores de qiskit
        elif self.base == "qiskit.aer":
            if params['backend']:
                self.backend = params['backend']
            else:
                raise Exception("Backend no encontrado")
            
            self.device= qml.device(self.base, backend=self.backend, wires=self.qubits)
        ##Simuladores de pennylane
        else:
            self.device= qml.device(self.base, wires=self.qubits)
        return
    
    
    def set_node(self, params) -> None:
        if "pattern" in params:
            self.pattern = params["pattern"]
        
        if params['repetitions']:
            self.repetition = params['repetitions']
        
        if params['interface']:
            self.interface= params['interface']
        
        if params['diff_method']:
            self.diff_method = params['diff_method']
        
        if params['interface'] == "jax" or params['interface'] == "jax-jit":
            node = qml.QNode(self.circuit, self.device, interface=self.interface, diff_method=self.diff_method)
            self.node = jax.jit(node)
        else:
            self.node = qml.QNode(self.circuit, self.device, interface=self.interface, diff_method=self.diff_method)
        return
    

    def get_state(self, theta):
        if self.interface == "jax" or self.interface == "jax-jit":
            node = qml.QNode(self.circuit_state, self.device, interface=self.interface, diff_method=self.diff_method)
            node = jax.jit(node)
        else:
            node = qml.QNode(self.circuit_state, self.device, interface=self.interface, diff_method=self.diff_method)
        return node(theta)